from __future__ import absolute_import, division, print_function

from contextlib import contextmanager

from mock import MagicMock

from glue import core
from glue.core.application_base import Application
from glue.tests.helpers import make_file


@contextmanager
def simple_catalog():
    """Context manager to create a temporary data file

    :param suffix: File suffix. string
    """
    with make_file(b'#a, b\n1, 2\n3, 4', '.csv') as result:
        yield result


def simple_session():
    collect = core.data_collection.DataCollection()
    hub = core.hub.Hub()
    result = core.Session(data_collection=collect, hub=hub,
                          application=MagicMock(Application),
                          command_stack=core.CommandStack())
    result.command_stack.session = result
    return result



from __future__ import absolute_import, division, print_function

from qtpy import compat
from glue.config import viewer_tool
from glue.viewers.common.qt.tool import Tool


@viewer_tool
class PythonExportTool(Tool):

    icon = 'glue_pythonsave'
    tool_id = 'save:python'
    action_text = 'Save Python script to reproduce plot'
    tool_tip = 'Save Python script to reproduce plot'

    def activate(self):

        filename, _ = compat.getsavefilename(parent=self.viewer, basedir="make_plot.py")

        if not filename:
            return

        if not filename.endswith('.py'):
            filename += '.py'

        self.viewer.export_as_script(filename)



from __future__ import absolute_import, division, print_function

from ..data_slice_widget import SliceWidget


class TestSliceWidget(object):

    def test_slice_center(self):
        s = SliceWidget(lo=0, hi=10)
        assert s.state.slice_center == 5

    def test_browse_slice(self):
        s = SliceWidget(lo=0, hi=10)
        assert s.state.slice_center == 5
        s.button_prev.click()
        assert s.state.slice_center == 4
        s.button_next.click()
        s.button_next.click()
        assert s.state.slice_center == 6
        s.button_first.click()
        assert s.state.slice_center == 0
        s.button_prev.click()
        assert s.state.slice_center == 10
        s.button_next.click()
        assert s.state.slice_center == 0
        s.button_last.click()
        assert s.state.slice_center == 10
        s.button_next.click()
        assert s.state.slice_center == 0
        s.button_prev.click()
        assert s.state.slice_center == 10
        s.button_prev.click()
        assert s.state.slice_center == 9

    def test_slice_world(self):

        s = SliceWidget(lo=0, hi=5, world=[1, 3, 5, 5.5, 8, 12])

        # Check switching between world and pixel coordinates
        s.state.slice_center = 0
        assert s.state.slider_label == '1.0'
        s.state.use_world = False
        assert s.state.slider_label == '0'
        s.state.slice_center = 3
        assert s.state.slider_label == '3'
        s.state.use_world = True
        assert s.state.slider_label == '5.5'

        # Round to nearest
        s.state.slider_label = '11'
        assert s.state.slice_center == 5
        assert s.state.slider_label == '12.0'

        # Make sure out of bound values work
        s.state.slider_label = '20'
        assert s.state.slice_center == 5
        assert s.state.slider_label == '12.0'
        s.state.slider_label = '-10'
        assert s.state.slice_center == 0
        assert s.state.slider_label == '1.0'

        # And disable world and try and set by pixel
        s.state.use_world = False
        s.state.slider_label = '4'
        assert s.state.slice_center == 4
        assert s.state.slider_label == '4'



from __future__ import absolute_import, division, print_function

from glue.external.echo import keep_in_sync, CallbackProperty
from glue.core.layer_artist import LayerArtistBase
from glue.viewers.common.state import LayerState
from glue.core.message import LayerArtistVisibilityMessage

__all__ = ['LayerArtist']


class LayerArtist(LayerArtistBase):

    zorder = CallbackProperty()
    visible = CallbackProperty()

    _layer_state_cls = LayerState

    def __init__(self, viewer_state, layer_state=None, layer=None):

        super(LayerArtist, self).__init__(layer)

        self._viewer_state = viewer_state

        self.layer = layer or layer_state.layer
        self.state = layer_state or self._layer_state_cls(viewer_state=viewer_state,
                                                          layer=self.layer)

        if self.state not in self._viewer_state.layers:
            self._viewer_state.layers.append(self.state)

        self.zorder = self.state.zorder
        self.visible = self.state.visible

        self._sync_zorder = keep_in_sync(self, 'zorder', self.state, 'zorder')
        self._sync_visible = keep_in_sync(self, 'visible', self.state, 'visible')

        self.state.add_callback('visible', self._on_visibility_change)

    def _on_visibility_change(self, *args):
        if self.state.layer is not None and self.state.layer.hub is not None:
            self.state.layer.hub.broadcast(LayerArtistVisibilityMessage(self))

    def __gluestate__(self, context):
        return dict(state=context.id(self.state))



from __future__ import absolute_import, division, print_function

from qtpy.QtCore import Qt
from qtpy import QtCore, QtWidgets
from glue.icons.qt import POINT_ICONS, symbol_icon
from glue.utils.qt import mpl_to_qt_color, qt_to_mpl_color


class ColorWidget(QtWidgets.QLabel):
    mousePressed = QtCore.Signal()

    def mousePressEvent(self, event):
        self.mousePressed.emit()
        event.accept()


class StyleDialog(QtWidgets.QDialog):

    """Dialog which edits the style of a layer (Data or Subset)

    Use via StyleDialog.edit_style(layer)
    """

    def __init__(self, layer, parent=None, edit_label=True):
        super(StyleDialog, self).__init__(parent)
        self.setWindowTitle("Style Editor")
        self.layer = layer
        self._edit_label = edit_label
        self._symbols = list(POINT_ICONS.keys())

        self._setup_widgets()
        self._connect()

    def _setup_widgets(self):
        self.layout = QtWidgets.QFormLayout()

        self.size_widget = QtWidgets.QSpinBox()
        self.size_widget.setMinimum(1)
        self.size_widget.setMaximum(40)
        self.size_widget.setValue(self.layer.style.markersize)

        self.label_widget = QtWidgets.QLineEdit()
        self.label_widget.setText(self.layer.label)
        self.label_widget.selectAll()

        self.color_widget = ColorWidget()
        self.color_widget.setStyleSheet('ColorWidget {border: 1px solid;}')
        color = self.layer.style.color
        color = mpl_to_qt_color(color, alpha=self.layer.style.alpha)
        self.set_color(color)

        self.okcancel = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok |
                                                   QtWidgets.QDialogButtonBox.Cancel)

        if self._edit_label:
            self.layout.addRow("Label", self.label_widget)
        self.layout.addRow("Color", self.color_widget)
        self.layout.addRow("Size", self.size_widget)

        self.layout.addWidget(self.okcancel)

        self.setLayout(self.layout)
        self.layout.setContentsMargins(6, 6, 6, 6)

    def _connect(self):
        self.color_widget.mousePressed.connect(self.query_color)
        self.okcancel.accepted.connect(self.accept)
        self.okcancel.rejected.connect(self.reject)
        self.setFocusPolicy(Qt.StrongFocus)

    def query_color(self, *args):
        color = QtWidgets.QColorDialog.getColor(self._color, self.color_widget,
                                                "",
                                                QtWidgets.QColorDialog.ShowAlphaChannel)
        if color.isValid():
            self.set_color(color)

    def color(self):
        return self._color

    def set_color(self, color):
        self._color = color
        pm = symbol_icon('o', color).pixmap(30, 30)
        self.color_widget.setPixmap(pm)

    def size(self):
        return self.size_widget.value()

    def label(self):
        return str(self.label_widget.text())

    def update_style(self):
        if self._edit_label:
            self.layer.label = self.label()
        self.layer.style.color = qt_to_mpl_color(self.color())
        self.layer.style.alpha = self.color().alpha() / 255.
        self.layer.style.markersize = self.size()

    @classmethod
    def edit_style(cls, layer):
        self = cls(layer)
        result = self.exec_()

        if result == self.Accepted:
            self.update_style()

    @classmethod
    def dropdown_editor(cls, item, pos, **kwargs):
        """
        Create a dropdown-style modal editor to edit the style of a
        given item

        :param item: Item with a .label and .style to edit
        :param pos: A QPoint to anchor the top-left corner of the dropdown at
        :param kwargs: Extra keywords to pass to StyleDialogs's constructor
        """
        self = cls(item, **kwargs)
        self.setWindowFlags(Qt.Tool | Qt.FramelessWindowHint)

        pos = self.mapFromGlobal(pos)
        self.move(pos)
        if self.exec_() == self.Accepted:
            self.update_style()


if __name__ == "__main__":

    from glue.utils.qt import get_qapp
    from glue.core import Data

    app = get_qapp()

    d = Data(label='data label', x=[1, 2, 3, 4])
    StyleDialog.edit_style(d)

    print("New layer properties")
    print(d.label)
    print('color: ', d.style.color)
    print('marker size: ', d.style.markersize)
    print('alpha ', d.style.alpha)



from __future__ import absolute_import, division, print_function

import os

import pkg_resources

__all__ = ['icon_path']


def icon_path(icon_name, icon_format='png'):
    """
    Return the absolute path to an icon

    Parameters
    ----------
    icon_name : str
       Name of icon, without extension or directory prefix
    icon_format : str, optional
        Can be either 'png' or 'svg'

    Returns
    -------
    path : str
      Full path to icon
    """

    icon_name += '.{0}'.format(icon_format)

    try:
        if pkg_resources.resource_exists('glue.icons', icon_name):
            return pkg_resources.resource_filename('glue.icons', icon_name)
        else:
            raise RuntimeError("Icon does not exist: %s" % icon_name)
    except NotImplementedError:  # workaround for mac app
        result = os.path.dirname(__file__)
        return os.path.join(result.replace('site-packages.zip', 'glue'),
                            icon_name)



from __future__ import absolute_import, division, print_function

from qtpy import QtCore, QtWidgets
from glue.core.edit_subset_mode import (NewMode, OrMode,
                                        AndNotMode, AndMode, XorMode,
                                        ReplaceMode)
from glue.app.qt.actions import action
from glue.utils import nonpartial, avoid_circular
from glue.utils.qt import update_combobox
from glue.core.message import EditSubsetMessage, SubsetMessage
from glue.core.hub import HubListener
from glue.external.six import string_types
from glue.icons.qt import layer_icon


class EditSubsetModeToolBar(QtWidgets.QToolBar, HubListener):

    def __init__(self, title="Subset mode toolbar", parent=None):
        super(EditSubsetModeToolBar, self).__init__(title, parent)

        self.subset_combo = QtWidgets.QComboBox()
        self.subset_combo.setMinimumContentsLength(10)

        spacer = QtWidgets.QWidget()
        spacer.setMinimumSize(10, 10)
        spacer.setSizePolicy(QtWidgets.QSizePolicy.Fixed,
                             QtWidgets.QSizePolicy.Preferred)

        self.addWidget(spacer)
        self.addWidget(QtWidgets.QLabel("Active Subset:"))
        self.addWidget(self.subset_combo)

        self._label_subset_mode = QtWidgets.QLabel("Mode:")
        self.addWidget(self._label_subset_mode)
        self.setIconSize(QtCore.QSize(16, 16))
        self._group = QtWidgets.QActionGroup(self)
        self._modes = {}
        self._add_actions()
        self._edit_subset_mode = self.parent()._session.edit_subset_mode
        self._modes[self._edit_subset_mode.mode].trigger()
        self._backup_mode = None

        spacer = QtWidgets.QWidget()
        spacer.setMinimumSize(20, 10)
        spacer.setSizePolicy(QtWidgets.QSizePolicy.Fixed,
                             QtWidgets.QSizePolicy.Preferred)

        self.parent()._hub.subscribe(self, EditSubsetMessage, handler=self._update_mode)
        self.parent()._hub.subscribe(self, SubsetMessage, handler=self._update_subset_combo)

        self._data_collection = self.parent().data_collection
        self._update_subset_combo()
        self.subset_combo.currentIndexChanged.connect(self._on_subset_combo_change)

        self._update_mode_visibility()

    def _update_subset_combo(self, msg=None):
        """
        Set up the combo listing the subsets.
        """

        # Prepare contents of combo box - we include a 'Create subset' item as
        # the last item
        labeldata = [(subset.label, subset) for subset in self._data_collection.subset_groups]
        labeldata.append(('None/Create New', None))

        # We now update the combo box, but we block the signals as we don't want
        # this to cause the current subset being edited to be modified.

        self.subset_combo.blockSignals(True)

        # The block_signals here is to prevent signals from being turned back
        # on inside update_combobox.
        update_combobox(self.subset_combo, labeldata, block_signals=False)
        self.subset_combo.setIconSize(QtCore.QSize(12, 12))
        for index, subset in enumerate(self._data_collection.subset_groups):
            self.subset_combo.setItemIcon(index, layer_icon(subset))

        self.subset_combo.blockSignals(False)

    def _update_mode_visibility(self):
        visible_modes = not self.subset_combo.currentIndex() == self.subset_combo.count() - 1
        if visible_modes:
            self._label_subset_mode.setText("Mode:")
        else:
            self._label_subset_mode.setText("(the next selection will create a subset)")
        for mode in self._modes.values():
            if isinstance(mode, QtWidgets.QAction):
                mode.setVisible(visible_modes)

    @avoid_circular
    def _on_subset_combo_change(self, event=None):
        """
        Update the EditSubsetMode when the subset combo changes.
        """

        subset = self.subset_combo.currentData()

        if subset is None:
            self._edit_subset_mode.edit_subset = []
        else:
            self._edit_subset_mode.edit_subset = [self.subset_combo.currentData()]
            if self._edit_subset_mode.mode is NewMode:
                self._edit_subset_mode.mode = ReplaceMode
        self._update_mode_visibility()

        # We now force the combo to be refreshed in case it included the
        # temporary 'Multiple subsets' entry which needs to be removed.
        self.subset_combo.blockSignals(True)
        self._update_subset_combo()
        self.subset_combo.blockSignals(False)

    @avoid_circular
    def _on_edit_subset_mode_change(self):
        """
        Update the subset combo when the EditSubsetMode changes.
        """

        # We block signals since we don't want to trigger a change in EditSubsetMode.

        self.subset_combo.blockSignals(True)

        edit_subset = self._edit_subset_mode.edit_subset

        if self._edit_subset_mode.mode is NewMode:
            index = self.subset_combo.count() - 1
        elif len(edit_subset) > 1:
            # We temporarily add an item - we remove this if the combo changes
            # again.
            self.subset_combo.insertItem(0, 'Multiple subsets')
            index = 0
        else:
            self._update_subset_combo()
            # In some cases, this can be called from _update_subset_combo before
            # EditSubsetMode has been updated, so the edit subset may still be
            # the old one and not exist in the list of subset groups.
            if edit_subset:
                if edit_subset[0] in self._data_collection.subset_groups:
                    index = self._data_collection.subset_groups.index(edit_subset[0])
                else:
                    index = self.subset_combo.count() - 1
            elif len(edit_subset) == 0:
                index = self.subset_combo.count() - 1

        self.subset_combo.setCurrentIndex(index)
        self.subset_combo.blockSignals(False)

        self._update_mode_visibility()

    def _make_mode(self, name, tip, icon, mode):

        def set_mode(mode):
            self._edit_subset_mode.mode = mode

        a = action(name, self, tip, icon)
        a.setCheckable(True)
        a.triggered.connect(nonpartial(set_mode, mode))
        self._group.addAction(a)
        self.addAction(a)
        self._modes[mode] = a
        label = name.split()[0].lower().replace('&', '')
        self._modes[label] = mode

    def _add_actions(self):
        self._make_mode("&Replace Mode", "Replace selection",
                        'glue_replace', ReplaceMode)
        self._make_mode("&Or Mode", "Add to selection",
                        'glue_or', OrMode)
        self._make_mode("&And Mode", "Set selection as intersection",
                        'glue_and', AndMode)
        self._make_mode("&Xor Mode", "Set selection as exclusive intersection",
                        'glue_xor', XorMode)
        self._make_mode("&Not Mode", "Remove from selection",
                        'glue_andnot', AndNotMode)

    def _update_mode(self, message):
        self.set_mode(message.mode)
        self._on_edit_subset_mode_change()

    def set_mode(self, mode):
        """Temporarily set the edit mode to mode
        :param mode: Name of the mode (Or, Not, And, Xor, Replace)
        :type mode: str
        """
        if mode == 'new' or mode is NewMode:
            if self._edit_subset_mode.mode is not NewMode:
                self._edit_subset_mode.mode = NewMode
            for act in self._modes.values():
                if isinstance(act, QtWidgets.QAction):
                    act.setChecked(False)
            return
        elif isinstance(mode, string_types):
            try:
                mode = self._modes[mode]  # label to mode class
            except KeyError:
                raise KeyError("Unrecognized mode: %s" % mode)

        self._backup_mode = self._backup_mode or self._edit_subset_mode.mode
        self._modes[mode].trigger()  # mode class to action

    def unset_mode(self):
        """Restore the mode to the state before set_mode was called"""
        mode = self._backup_mode
        self._backup_mode = None
        if mode:
            self._modes[mode].trigger()



from __future__ import absolute_import, division, print_function

import warnings
from os.path import basename
from collections import OrderedDict

from glue.core.coordinates import coordinates_from_header, WCSCoordinates
from glue.core.data import Component, Data
from glue.config import data_factory, qglue_parser
from glue.external.six import string_types

__all__ = ['is_fits', 'fits_reader', 'is_casalike', 'casalike_cube']


def is_fits(filename):
    from astropy.io import fits
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            with fits.open(filename, ignore_missing_end=True, mode='denywrite'):
                return True
    except IOError:
        return False


@data_factory(
    label='FITS file',
    identifier=is_fits,
    priority=100,
)
def fits_reader(source, auto_merge=False, exclude_exts=None, label=None):
    """
    Read in all extensions from a FITS file.

    Parameters
    ----------
    source: str or HDUList
        The pathname to the FITS file.
        If an HDUList is passed in, simply use that.

    auto_merge: bool
        Merge extensions that have the same shape
        and only one has a defined WCS.

    exclude_exts: [hdu, ] or [index, ]
        List of HDU's to exclude from reading.
        This can be a list of HDU's or a list
        of HDU indexes.
    """

    from astropy.io import fits
    from astropy.table import Table

    exclude_exts = exclude_exts or []

    if isinstance(source, fits.hdu.hdulist.HDUList):
        hdulist = source
        close_hdulist = False
    else:
        hdulist = fits.open(source, ignore_missing_end=True, mode='denywrite')
        hdulist.verify('fix')
        close_hdulist = True

    groups = OrderedDict()
    extension_by_shape = OrderedDict()

    if label is not None:

        label_base = label

    else:

        hdulist_name = hdulist.filename()
        if hdulist_name is None:
            hdulist_name = "HDUList"

        label_base = basename(hdulist_name).rpartition('.')[0]

        if not label_base:
            label_base = basename(hdulist_name)

    # Create a new image Data.
    def new_data(suffix=True):
        if suffix:
            label = '{0}[{1}]'.format(label_base, hdu_name)
        else:
            label = label_base
        data = Data(label=label)
        data.coords = coords

        # We need to be careful here because some header values are special
        # objects that we should convert to strings
        for key, value in hdu.header.items():
            if (key == 'COMMENT' or key == 'HISTORY'):
                if key not in data.meta:
                    data.meta[key] = [str(value)]
                else:
                    data.meta[key].append(str(value))
            elif isinstance(value, string_types) or isinstance(value, (int, float, bool)):
                data.meta[key] = value
            else:
                data.meta[key] = str(value)

        groups[hdu_name] = data
        extension_by_shape[shape] = hdu_name
        return data

    for extnum, hdu in enumerate(hdulist):
        hdu_name = hdu.name if hdu.name else "HDU{0}".format(extnum)
        if (hdu.data is not None and
                hdu.data.size > 0 and
                hdu_name not in exclude_exts and
                extnum not in exclude_exts):
            if is_image_hdu(hdu):
                shape = hdu.data.shape
                coords = coordinates_from_header(hdu.header)
                units = hdu.header.get('BUNIT')
                if not auto_merge or has_wcs(coords):
                    data = new_data(suffix=len(hdulist) > 1)
                else:
                    try:
                        data = groups[extension_by_shape[shape]]
                    except KeyError:
                        data = new_data(suffix=len(hdulist) > 1)
                component = Component.autotyped(hdu.data, units=units)
                data.add_component(component=component,
                                   label=hdu_name)
            elif is_table_hdu(hdu):
                # Loop through columns and make component list
                table = Table.read(hdu, format='fits')
                label = '{0}[{1}]'.format(label_base, hdu_name)
                data = Data(label=label)
                groups[hdu_name] = data
                for column_name in table.columns:
                    column = table[column_name]
                    if column.ndim != 1:
                        warnings.warn("Dropping column '{0}' since it is not 1-dimensional".format(column_name))
                        continue
                    component = Component.autotyped(column, units=column.unit)
                    data.add_component(component=component,
                                       label=column_name)

    if close_hdulist:
        hdulist.close()

    return [groups[idx] for idx in groups]


# Utilities

def is_image_hdu(hdu):
    from astropy.io.fits.hdu import PrimaryHDU, ImageHDU, CompImageHDU
    return isinstance(hdu, (PrimaryHDU, ImageHDU, CompImageHDU))


def is_table_hdu(hdu):
    from astropy.io.fits.hdu import TableHDU, BinTableHDU
    return isinstance(hdu, (TableHDU, BinTableHDU))


def has_wcs(coords):
    return (isinstance(coords, WCSCoordinates) and
            any(axis['coordinate_type'] is not None
                for axis in coords.wcs.get_axis_types()))


def is_casalike(filename, **kwargs):
    """
    Check if a FITS file is a CASA like cube,
    with (P, P, V, Stokes) layout
    """
    from astropy.io import fits

    if not is_fits(filename):
        return False
    with fits.open(filename, ignore_missing_end=True, mode='denywrite') as hdulist:
        if len(hdulist) != 1:
            return False
        if hdulist[0].header['NAXIS'] != 4:
            return False

        from astropy.wcs import WCS
        w = WCS(hdulist[0].header)

    ax = [a.get('coordinate_type') for a in w.get_axis_types()]
    return ax == ['celestial', 'celestial', 'spectral', 'stokes']


@data_factory(label='CASA PPV Cube', identifier=is_casalike, deprecated=True)
def casalike_cube(filename, **kwargs):
    """
    This provides special support for 4D CASA FITS - like cubes,
    which have 2 spatial axes, a spectral axis, and a stokes axis
    in that order.

    Each stokes cube is split out as a separate component
    """
    from astropy.io import fits

    result = Data()

    if 'ignore_missing_end' not in kwargs:
        kwargs['ignore_missing_end'] = True

    with fits.open(filename, mode='denywrite', **kwargs) as hdulist:
        array = hdulist[0].data
        header = hdulist[0].header
    result.coords = coordinates_from_header(header)
    for i in range(array.shape[0]):
        units = header.get('BUNIT')
        component = Component.autotyped(array[[i]], units=units)
        result.add_component(component, label='STOKES %i' % i)
    return result


try:
    from astropy.io.fits import HDUList
except ImportError:
    pass
else:
    # Put HDUList parser before list parser
    @qglue_parser(HDUList, priority=100)
    def _parse_data_hdulist(data, label):
        from glue.core.data_factories.fits import fits_reader
        return fits_reader(data, label=label)



from __future__ import absolute_import, division, print_function

from qtpy import QtWidgets

__all__ = ['pick_item', 'pick_class', 'get_text', 'CenteredDialog']


def pick_item(items, labels, title="Pick an item", label="Pick an item",
              default=None):
    """
    Prompt the user to choose an item

    Returns the selected item, or `None`

    Parameters
    ----------
    items : iterable
        Items to choose (can be any Python object)
    labels : iterables
        Labels for the items to choose
    title : str, optional
        The title of the dialog
    label : str, optional
        The prompt message
    """

    if default in items:
        current = items.index(default)
    else:
        current = 0

    choice, isok = QtWidgets.QInputDialog.getItem(None, title, label,
                                                  labels, current=current,
                                                  editable=False)
    if isok:
        index = labels.index(str(choice))
        return items[index]


def pick_class(classes, sort=False, **kwargs):
    """
    Prompt the user to pick from a list of classes using Qt

    This is the same as `pick_item`, but the labels are automatically determined
    from the classes using the LABEL attribute, and if not set, then the
    __name__.

    Returns the class that was selected, or `None`

    Parameters
    ----------
    classes : iterable
        The classes to choose from
    title : str, optional
        The title of the dialog
    label : str, optional
        The prompt message
    """
    def _label(c):
        try:
            return c.LABEL
        except AttributeError:
            return c.__name__

    if sort:
        classes = sorted(classes, key=lambda x: _label(x))
    choices = [_label(c) for c in classes]
    return pick_item(classes, choices, **kwargs)


def get_text(title='Enter a label', default=None):
    """
    Prompt the user to enter text using Qt

    Returns the text the user typed, or `None`

    Parameters
    ----------
    title : str
        The prompt message and widget title.
    default : str
        The default text to show in the prompt.
    """
    result, isok = QtWidgets.QInputDialog.getText(None, title, title, text=default)
    if isok:
        return str(result)


class CenteredDialog(QtWidgets.QDialog):
    """
    A dialog that is centered on the screen.
    """

    def center(self):
        # Adapted from StackOverflow
        # https://stackoverflow.com/questions/20243637/pyqt4-center-window-on-active-screen
        frameGm = self.frameGeometry()
        screen = QtWidgets.QApplication.desktop().screenNumber(QtWidgets.QApplication.desktop().cursor().pos())
        centerPoint = QtWidgets.QApplication.desktop().screenGeometry(screen).center()
        frameGm.moveCenter(centerPoint)
        self.move(frameGm.topLeft())



from __future__ import absolute_import, division, print_function

import platform
from qtpy import QtCore, QtGui, QtWidgets

__all__ = ['get_qapp', 'fix_tab_widget_fontsize']

qapp = None


def get_qapp(icon_path=None):

    global qapp

    qapp = QtWidgets.QApplication.instance()

    if qapp is None:

        # NOTE: plugins that need WebEngine may complain that QtWebEngineWidgets
        # needs to be imported before QApplication is constructed, but this can
        # cause segmentation faults to crop up under certain conditions, so we
        # don't do it here and instead ask that the plugins do it in their
        # main __init__.py (which should get executed before glue is launched).

        qapp = QtWidgets.QApplication([''])
        qapp.setQuitOnLastWindowClosed(True)

        if icon_path is not None:
            qapp.setWindowIcon(QtGui.QIcon(icon_path))

        if platform.system() == 'Darwin':
            # On Mac, the fonts are generally too large compared to other
            # applications, so we reduce the default here. In future we should
            # make this a setting in the system preferences.
            size_offset = 2
        else:
            # On other platforms, we reduce the font size by 1 point to save
            # space too. Again, this should probably be a global setting.
            size_offset = 1

        font = qapp.font()
        font.setPointSize(font.pointSize() - size_offset)
        qapp.setFont(font)

    # Make sure we use high resolution icons for HDPI displays.
    qapp.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps)

    return qapp


def fix_tab_widget_fontsize(tab_widget):
    """
    Because of a bug in Qt, tab titles on MacOS X don't have the right font size
    """
    if platform.system() == 'Darwin':
        app = get_qapp()
        app_font = app.font()
        tab_widget.setStyleSheet('font-size: {0}px'.format(app_font.pointSize()))



from __future__ import absolute_import, division, print_function

from glue.utils import defer_draw, decorate_all_methods

from glue.viewers.matplotlib.qt.data_viewer import MatplotlibDataViewer
from glue.viewers.profile.qt.layer_style_editor import ProfileLayerStyleEditor
from glue.viewers.profile.layer_artist import ProfileLayerArtist
from glue.viewers.profile.qt.options_widget import ProfileOptionsWidget
from glue.viewers.profile.state import ProfileViewerState

from glue.viewers.common.qt import toolbar_mode  # noqa
from glue.viewers.profile.qt.profile_tools import ProfileAnalysisTool  # noqa
from glue.viewers.profile.viewer import MatplotlibProfileMixin

__all__ = ['ProfileViewer']


@decorate_all_methods(defer_draw)
class ProfileViewer(MatplotlibProfileMixin, MatplotlibDataViewer):

    LABEL = '1D Profile'
    _layer_style_widget_cls = ProfileLayerStyleEditor
    _state_cls = ProfileViewerState
    _options_cls = ProfileOptionsWidget
    _data_artist_cls = ProfileLayerArtist
    _subset_artist_cls = ProfileLayerArtist

    large_data_size = 1e8

    allow_duplicate_data = True

    tools = ['select:xrange', 'profile-analysis']

    def __init__(self, session, parent=None, state=None):
        MatplotlibDataViewer.__init__(self, session, parent=parent, state=state)
        MatplotlibProfileMixin.setup_callbacks(self)



# pylint: disable=I0011,W0613,W0201,W0212,E1101,E1103

from __future__ import absolute_import, division, print_function

import numpy as np
from mock import MagicMock, patch

from glue.core import Data, DataCollection
from glue.app.qt import GlueApplication

from glue.core.tests.util import simple_session
from glue.viewers.histogram.qt import HistogramViewer
from glue.viewers.image.qt import ImageViewer
from glue.viewers.scatter.qt import ScatterViewer
from glue.utils.qt import get_qapp

# TODO: We should maybe consider running these tests for all
# registered Qt viewers.


class BaseTestDataViewer(object):

    ndim = 1

    def test_unregister_on_close(self):
        session = simple_session()
        hub = session.hub

        w = self.widget_cls(session)
        w.register_to_hub(hub)
        with patch.object(w, 'unregister') as unregister:
            w.close()
        unregister.assert_called_once_with(hub)

    def test_single_draw_call_on_create(self):
        d = Data(x=np.random.random((2,) * self.ndim))
        dc = DataCollection([d])
        app = GlueApplication(dc)

        try:
            from glue.viewers.matplotlib.qt.widget import MplCanvas
            draw = MplCanvas.draw
            MplCanvas.draw = MagicMock()

            app.new_data_viewer(self.widget_cls, data=d)

            # each Canvas instance gives at most 1 draw call
            selfs = [c[0][0] for c in MplCanvas.draw.call_arg_list]
            assert len(set(selfs)) == len(selfs)
        finally:
            MplCanvas.draw = draw
        app.close()

    def test_close_on_last_layer_remove(self):

        # regression test for 391

        # Note: processEvents is needed for things to work correctly with PySide2
        qtapp = get_qapp()

        d1 = Data(x=np.random.random((2,) * self.ndim))
        d2 = Data(y=np.random.random((2,) * self.ndim))
        dc = DataCollection([d1, d2])
        app = GlueApplication(dc)
        w = app.new_data_viewer(self.widget_cls, data=d1)
        w.add_data(d2)
        qtapp.processEvents()
        assert len(app.viewers[0]) == 1
        dc.remove(d1)
        qtapp.processEvents()
        assert len(app.viewers[0]) == 1
        dc.remove(d2)
        qtapp.processEvents()
        assert len(app.viewers[0]) == 0
        app.close()

    def test_viewer_size(self, tmpdir):

        # regression test for #781
        # viewers were not restored with the right size

        d1 = Data(x=np.random.random((2,) * self.ndim))
        d2 = Data(x=np.random.random((2,) * self.ndim))
        dc = DataCollection([d1, d2])
        app = GlueApplication(dc)
        w = app.new_data_viewer(self.widget_cls, data=d1)
        w.viewer_size = (300, 400)

        filename = tmpdir.join('session.glu').strpath
        app.save_session(filename, include_data=True)

        app2 = GlueApplication.restore_session(filename)

        for viewer in app2.viewers:
            assert viewer[0].viewer_size == (300, 400)

        app.close()
        app2.close()


class TestDataViewerScatter(BaseTestDataViewer):
    widget_cls = ScatterViewer


class TestDataViewerImage(BaseTestDataViewer):
    ndim = 2
    widget_cls = ImageViewer


class TestDataViewerHistogram(BaseTestDataViewer):
    widget_cls = HistogramViewer



from __future__ import absolute_import, division, print_function

from qtpy import QtWidgets
from ..helpers import update_combobox


def test_update_combobox():
    combo = QtWidgets.QComboBox()
    update_combobox(combo, [('a', 1), ('b', 2)])
    update_combobox(combo, [('c', 3)])


def test_update_combobox_indexchanged():

    # Regression test for bug that caused currentIndexChanged to not be
    # emitted if the new index happened to be the same as the old one but the
    # label data was different.

    class MyComboBox(QtWidgets.QComboBox):

        def __init__(self, *args, **kwargs):
            self.change_count = 0
            super(MyComboBox, self).__init__(*args, **kwargs)
            self.currentIndexChanged.connect(self.changed)

        def changed(self):
            self.change_count += 1

    combo = MyComboBox()
    update_combobox(combo, [('a', 1), ('b', 2)])
    update_combobox(combo, [('c', 3)])

    assert combo.change_count == 2
    assert combo.currentIndex() == 0

    combo = MyComboBox()
    update_combobox(combo, [('a', 1), ('b', 2)])
    update_combobox(combo, [('a', 1), ('b', 3)])
    update_combobox(combo, [('a', 3), ('b', 1)])

    assert combo.change_count == 3
    assert combo.currentIndex() == 1



import numpy as np
from astropy import units as u
from astropy.wcs import WCSSUB_CELESTIAL, WCSSUB_SPECTRAL


def get_spatial_scale(wcs, assert_square=True):

    # Code adapted from APLpy

    wcs = wcs.sub([WCSSUB_CELESTIAL])
    cdelt = np.matrix(wcs.wcs.get_cdelt())
    pc = np.matrix(wcs.wcs.get_pc())
    scale = np.array(cdelt * pc)

    if assert_square:
        try:
            np.testing.assert_almost_equal(abs(cdelt[0,0]), abs(cdelt[0,1]))
            np.testing.assert_almost_equal(abs(pc[0,0]), abs(pc[1,1]))
            np.testing.assert_almost_equal(abs(scale[0,0]), abs(scale[0,1]))
        except AssertionError:
            raise ValueError("Non-square pixels.  Please resample data.")

    return abs(scale[0,0]) * u.Unit(wcs.wcs.cunit[0])


def get_spectral_scale(wcs):

    # Code adapted from APLpy

    wcs = wcs.sub([WCSSUB_SPECTRAL])
    cdelt = np.matrix(wcs.wcs.get_cdelt())
    pc = np.matrix(wcs.wcs.get_pc())
    scale = np.array(cdelt * pc)

    return abs(scale[0,0]) * u.Unit(wcs.wcs.cunit[0])


def sanitize_wcs(mywcs):
    pc = np.matrix(mywcs.wcs.get_pc())
    if (pc[:,2].sum() != pc[2,2] or pc[2,:].sum() != pc[2,2]):
        raise ValueError("Non-independent 3rd axis.")
    axtypes = mywcs.get_axis_types()
    if ((axtypes[0]['coordinate_type'] != 'celestial' or
         axtypes[1]['coordinate_type'] != 'celestial' or
         axtypes[2]['coordinate_type'] != 'spectral')):
        cunit3 = mywcs.wcs.cunit[2]
        ctype3 = mywcs.wcs.ctype[2]
        if cunit3 != '':
            cunit3 = u.Unit(cunit3)
            if cunit3.is_equivalent(u.m/u.s):
                mywcs.wcs.ctype[2] = 'VELO'
            elif cunit3.is_equivalent(u.Hz):
                mywcs.wcs.ctype[2] = 'FREQ'
            elif cunit3.is_equivalent(u.m):
                mywcs.wcs.ctype[2] = 'WAVE'
            else:
                raise ValueError("Could not determine type of 3rd axis.")
        elif ctype3 != '':
            if 'VELO' in ctype3:
                mywcs.wcs.ctype[2] = 'VELO'
            elif 'FELO' in ctype3:
                mywcs.wcs.ctype[2] = 'VELO-F2V'
            elif 'FREQ' in ctype3:
                mywcs.wcs.ctype[2] = 'FREQ'
            elif 'WAVE' in ctype3:
                mywcs.wcs.ctype[2] = 'WAVE'
            else:
                raise ValueError("Could not determine type of 3rd axis.")
        else:
            raise ValueError("Cube axes not in expected orientation: PPV")
    return mywcs


def get_wcs_system_frame(wcs):
    """TODO: move to astropy.wcs.utils"""
    ct = wcs.sub([WCSSUB_CELESTIAL]).wcs.ctype
    if 'GLON' in ct[0]:
        from astropy.coordinates import Galactic
        return Galactic
    elif 'RA' in ct[0]:
        from astropy.coordinates import ICRS
        return ICRS
    else:
        raise ValueError("Unrecognized coordinate system")



import numpy as np
from .geometry import path
from astropy import coordinates
from astropy import units as u
import re

csystems = {'galactic':coordinates.Galactic,
            'fk5':coordinates.FK5,
            'fk4':coordinates.FK4,
            'icrs':coordinates.ICRS}
cel_systems = ['fk5','fk4','icrs']
# ecliptic, detector, etc. not supported (because I don't know what they mean)
# (or with ecliptic, how to deal with them)
all_systems = cel_systems+['galactic','image','physical']

class SimpleRegion(object):
    def __init__(self, coord_list, coord_format, name):
        self.name = name
        self.coord_format = coord_format
        self.coord_list = coord_list

    def __repr__(self):
        return "Region: {0}, {1}, {2}".format(self.name, self.coord_list,
                                              self.coord_format)


valid_regions = ['line', 'segment', 'vector']
valid_region_re = [re.compile("^"+n) for n in valid_regions]

def simple_region_parser(regionstring, coord_format):
    rs = regionstring.lstrip("# ")

    rtype = None
    for rt, rre in zip(valid_regions, valid_region_re):
        if rre.search(rs):
            rtype = rt
            break

    if rtype is None:
        # not a usable region
        return

    coordre = re.compile("^[a-z]*\((.*)\)")
    coord_list = coordre.findall(rs)
    if len(coord_list) != 1:
        raise ValueError("Invalid region")

    coords = coord_list[0].split(",")

    outcoords = []
    for ii,cs in enumerate(coords):
        if coord_format in csystems:
            if ":" in cs:
                # sexagesimal
                if coord_format in cel_systems and ii % 2 == 0:
                    # odd, celestial = RA = hours
                    crd = coordinates.Angle(cs, unit=u.hour)
                else:
                    crd = coordinates.Angle(cs, unit=u.deg)
            else:
                try:
                    # if it's a float, it's in degrees
                    crd = float(cs) * u.deg
                except ValueError:
                    crd = coordinates.Angle(cs)
        else:
            # assume pixel units
            crd = float(cs)
        outcoords.append(crd)

    reg = SimpleRegion(coord_list=outcoords, coord_format=coord_format,
                       name=rtype)

    return reg

def load_regions_file(rfile):
    with open(rfile,'r') as fh:
        lines = fh.readlines()
    return load_regions_stringlist(lines)

def load_regions_stringlist(lines):

    coord_format = None
    for line in lines:
        if line.strip() in all_systems:
            coord_format = line.strip()
            break
    if coord_format is None:
        raise ValueError("No valid coordinate format found.")

    regions_ = [simple_region_parser(line, coord_format) for line in lines]
    regions = [r for r in regions_ if r is not None]
    
    return regions


def line_to_path(region):
    """
    Convert a line or segment to a path
    """

    l,b = None,None

    endpoints = []

    for x in region.coord_list:
        if l is None:
            l = x.to(u.deg).value
        else:
            b = x.to(u.deg).value
            if l is not None and b is not None:
                endpoints.append((l,b))
                l,b = None,None
            else:
                raise ValueError("unmatched l,b")

    lbarr = np.array(endpoints)
    C = csystems[region.coord_format](lbarr[:,0]*u.deg, lbarr[:,1]*u.deg)

    # TODO: add widths for projection

    p = path.Path(C)

    return p

def vector_to_path(vector_region):
    """
    Convert a vector region to a path

    # vector(48.944348,-0.36432694,485.647",124.082) vector=1
    """

    x,y = vector_region.coord_list[:2]
    length = vector_region.coord_list[2]
    angle = vector_region.coord_list[3]
    
    C1 = csystems[vector_region.coord_format](x, y)
    dx,dy = length * np.cos(angle), length * np.sin(angle)
    # -dx because we're in the flippy coordsys
    C2 = csystems[vector_region.coord_format](C1.lonangle - dx, C1.latangle + dy)

    C = csystems[vector_region.coord_format]([C1.lonangle,C2.lonangle],
                                             [C1.latangle,C2.latangle])

    p = path.Path(C)

    return p

region_converters = {'line':line_to_path, 'segment':line_to_path,
                     'vector':vector_to_path}

def paths_from_regfile(regfile):
    """
    Given a ds9 region file, extract pv diagrams for each:
        group of points [NOT IMPLEMENTED]
        panda [NOT IMPLEMENTED]
        vector [NOT IMPLEMENTED]
        segment [NOT IMPLEMENTED]
        group of lines [NOT IMPLEMENTED]
    """
    #import pyregion
    #regions = pyregion.open(regfile)
    regions = load_regions_file(regfile)
    return paths_from_regions(regions)

def paths_from_regions(regions):
    paths = [region_converters[r.name](r)
             for r in regions
             if r.name in region_converters]
    return paths



from __future__ import absolute_import, division, print_function

import weakref
from glue.config import viewer_tool
from glue.viewers.common.qt.tool import Tool
from glue.core.qt.dialogs import info, warn


@viewer_tool
class ProfileViewerTool(Tool):

    icon = 'glue_spectrum'
    tool_id = 'profile-viewer'

    def __init__(self, viewer):
        super(ProfileViewerTool, self).__init__(viewer)
        self.profile_viewers = []

    def activate(self):

        if any(x() is not None for x in self.profile_viewers):

            proceed = warn('A profile viewer was already created',
                           'Do you really want to create a new one?',
                           default='Cancel', setting='show_warn_profile_duplicate')

            if not proceed:
                return

        else:

            proceed = info('Creating a profile viewer',
                           'Note: profiles are '
                           'computed from datasets and subsets collapsed along all but one '
                           'dimension. To view the profile of part of the data, once you '
                           'click OK you can draw and update a subset in the current '
                           'image viewer and the profile will update accordingly.', setting='show_info_profile_open')

            if not proceed:
                return

        from glue.viewers.profile.qt import ProfileViewer
        profile_viewer = self.viewer.session.application.new_data_viewer(ProfileViewer)
        any_added = False
        for data in self.viewer.session.data_collection:
            if data in self.viewer._layer_artist_container:
                result = profile_viewer.add_data(data)
                any_added = any_added or result
        if any_added:
            if self.viewer.state.reference_data in profile_viewer._layer_artist_container:
                profile_viewer.state.reference_data = self.viewer.state.reference_data
            self.profile_viewers.append(weakref.ref(profile_viewer))
        else:
            profile_viewer.close()



from __future__ import absolute_import, division, print_function

import os

from qtpy import QtWidgets

from glue.external.echo.qt import autoconnect_callbacks_to_qt
from glue.utils.qt import load_ui, fix_tab_widget_fontsize
from glue.viewers.matplotlib.state import MatplotlibDataViewerState

__all__ = ['HistogramOptionsWidget']


class HistogramOptionsWidget(QtWidgets.QWidget):

    def __init__(self, viewer_state, session, parent=None):

        super(HistogramOptionsWidget, self).__init__(parent=parent)

        self.ui = load_ui('options_widget.ui', self,
                          directory=os.path.dirname(__file__))

        fix_tab_widget_fontsize(self.ui.tab_widget)

        autoconnect_callbacks_to_qt(viewer_state, self.ui)
        autoconnect_callbacks_to_qt(viewer_state, self.ui.axes_editor.ui)

        self.viewer_state = viewer_state

        viewer_state.add_callback('x_att', self._update_attribute)

        self.session = session
        self.ui.axes_editor.button_apply_all.clicked.connect(self._apply_all_viewers)

    def _apply_all_viewers(self):
        for tab in self.session.application.viewers:
            for viewer in tab:
                if isinstance(viewer.state, MatplotlibDataViewerState):
                    viewer.state.update_axes_settings_from(self.viewer_state)

    def _update_attribute(self, *args):
        # If at least one of the components is categorical or a date, disable log button
        log_enabled = 'categorical' not in self.viewer_state.x_kinds
        self.ui.bool_x_log.setEnabled(log_enabled)
        self.ui.bool_x_log_.setEnabled(log_enabled)
        if not log_enabled:
            self.ui.bool_x_log.setChecked(False)
            self.ui.bool_x_log_.setChecked(False)
